module type INT_SET =
  sig
    type set
    val empty : set
    val isEmpty : set -> bool 
    val insert :  set -> int -> set
    val contains : set -> int -> bool
end;;

module IntSet : INT_SET =
  struct
    type set = Empty | Ins of int * set 
    let empty = Empty 
    let isEmpty s = (s = Empty)
    let insert s i = Ins(i,s)
    let rec contains s i =
      match s with 
       | Empty -> false 
       | Ins(j,r) -> if i = j then true   
                     else contains r i 
  end;;

module IntSetBST : INT_SET =
  struct
    type set = Tip | Bin of int * set * set
    let empty = Tip
    let isEmpty s = s = Tip
    let rec insert s i = 
     match s with 
        Tip -> Bin(i,Tip, Tip)
      | Bin(j,l,r) -> 
        if i = j then s 
        else if i < j then Bin(j,insert l i,r)
        else Bin(j,l,insert r i)
    let rec contains s i =
      match s with
	  Tip -> false
	| Bin(j,l,r) ->
	  if i = j then true
	  else if i < j then contains l i
	  else contains r i
  end
    
module IntSetF =
struct
  type set = bool * (int -> bool)
  let empty = (true, fun x-> false) 
  let isEmpty (e,_) = e 
  let insert (_,f) i =
    (false, if f i then f else fun j -> i = j || f j)
  let contains (_,f) i = f i
end;;

let e = IntSetF.empty;;
let one = IntSetF.insert e 1;;
let onetwo = IntSetF.insert one 2;;
true = IntSetF.contains  onetwo 2;;
false = IntSetF.contains onetwo 3;;
