(* Author Niki Vazou, April 2017                                      *)
(* Object Implementations in Ocaml                                    *)
(* From "On Understanding Data Abstractions, Revisited" by W. R. Cook *)

(* Fixpoints / Y combinator *)

(* Code taken from http://rosettacode.org/wiki/Y_combinator *)
(* val unroll : 'a mu -> 'a mu -> 'a = <fun>
  val fix : (('a -> 'b) -> 'a -> 'b) -> 'a -> 'b = <fun>
*)

type 'a mu = Roll of ('a mu -> 'a);;  (* this is needed to treat the same value as both having type 'a and type 'a -> 'b *)
let unroll (Roll x) = x;; 
 
let fix = fun f -> (fun x a -> f (unroll x x) a) (Roll (fun x a -> f (unroll x x) a));;
 
(* Object Interface *)

type iSet = 
	{isEmpty  : bool;
	 contains : int -> bool;
	 insert   : int -> iSet;
	}

(* Object Definitions *)

let rec insert_obj s n = 
 if s.contains n then s else 
 (fix (fun this _ ->  {
 	isEmpty  = false;
 	contains = (fun i -> i == n || s.contains i);
 	insert   = (fun i -> insert_obj (this ()) i);
 })) ()


let empty_obj _ = 
 (fix (fun this _ ->  {
 	isEmpty  = true;
 	contains = (fun i -> false);
 	insert   = (fun i -> insert_obj (this ()) i);
 })) ()


let union_obj s1 s2 = 
 (fix (fun this _ ->  {
 	isEmpty  = (s1.isEmpty && s2.isEmpty);
 	contains = (fun i -> s1.contains i || s2.contains i);
 	insert   = (fun i -> insert_obj (this ()) i);
 })) ()


let even_obj _ = 
 (fix (fun this _ ->  {
 	isEmpty  = false;
 	contains = (fun i -> (i mod 2) = 0);
 	insert   = (fun i -> insert_obj (this ()) i);
 })) ()


(* Tests *)

let ex_obj = union_obj (even_obj ()) ((empty_obj ()).insert 3);;
let ex_contains n = ex_obj.contains(n)





